import numpy as np

from StructuralCausalModels.directed_graph import DirectedGraph
from StructuralCausalModels.graph import InvalidAdjacencyMatrix


class DirectedAcyclicGraph(DirectedGraph):

    def __init__(self, name, adjacency_matrix):

        if not DirectedAcyclicGraph.validate_dag_adjacency_matrix(
                adjacency_matrix):
            msg = 'Adjacency matrix provided not valid for a DAG.'
            raise InvalidAdjacencyMatrix(msg)

        super().__init__(name, adjacency_matrix)

    @staticmethod
    def validate_dag_adjacency_matrix(matrix, atol=1e-6):
        """
        Checks that a matrix is a valid adjacency matrix for a directed acyclic
        graph. Uses the characterisation of acyclicity established by D. Wei, T.
        Gao and Y. Yu in "DAGs with No Fears : A Closer Look at Continuous
        Optimization for Learning Bayesian Networks" (2020) : a directed graph is
        acyclic if and only if its adjacency matrix is nilpotent.

        Parameters
        ----------
        matrix : array_like
            The matrix to check.
        atol : float
            The absolute tolerance used to check that the eigenvalues are all equal
            to 0.

        Returns
        -------
        bool
            Whether the matrix to check is a valid adjacency matrix for a directed
            acyclic graph.
        """
        if not DirectedGraph.validate_directed_graph_adjacency_matrix(matrix):
            return False

        eigenvalues = np.linalg.eigvals(matrix)
        comparand = np.zeros_like(eigenvalues)

        return np.allclose(eigenvalues, comparand, atol=atol)



import pytest
import numpy as np

from scipy.stats import randint

from StructuralCausalModels.structural_equation import StructuralEquation
from StructuralCausalModels.structural_causal_model import StructuralCausalModel, \
    InvalidIntervention

_constant_0 = 1
_constant_1 = 2
_constant_2 = -4
_constant_3 = 0


@pytest.fixture
def nb_samples():

    return 1000


@pytest.fixture
def deterministic_scm():
    """
    Defines an SCM in which all th exogenous variables are constant. This will
    allow a degree of checks on the construction of SCMs via the class defined
    to be carried out.
    """
    # X_0 = epsilon_0, where epsilon_0 = constant_0
    def f_0(u, *args):
        return u
    constant_0 = _constant_0
    epsilon_0 = randint(low=constant_0, high=(constant_0 + 1))
    equation_0 = StructuralEquation(0, [], epsilon_0, f_0)

    # X_3 = X_0 +  epsilon_3, where epsilon_3 = constant_3
    def f_3(u, x):
        return (u + np.power(x, 1)).flatten()
    constant_3 = _constant_3
    epsilon_3 = randint(low=constant_3, high=(constant_3 + 1))
    equation_3 = StructuralEquation(3, [0], epsilon_3, f_3)

    # X_1 = X_0 - X_3 + epsilon_1, where epsilon_1 = constant_1
    def f_1(u, x, y):
        return u + x - y
    constant_1 = _constant_1
    epsilon_1 = randint(low=constant_1, high=(constant_1 + 1))
    equation_1 = StructuralEquation(1, [0, 3], epsilon_1, f_1)

    # X_2 = X_0 - X_3 + X_1 + epsilon_2, where epsilon_2 = constant_2
    def f_2(u, x, y, z):
        return u + x - y + z
    constant_2 = _constant_2
    epsilon_2 = randint(low=constant_2, high=(constant_2 + 1))
    equation_2 = StructuralEquation(2, [0, 3, 1], epsilon_2, f_2)

    equations = [equation_0, equation_3, equation_1, equation_2]

    scm = StructuralCausalModel('deterministic_scm', len(equations), equations)

    return scm


@pytest.fixture
def deterministic_scm_data_generation_function(deterministic_scm, nb_samples):

    data = np.zeros(shape=(nb_samples, deterministic_scm.nb_var))
    data[:, 0] = _constant_0
    data[:, 3] = data[:, 0] + _constant_3
    data[:, 1] = data[:, 0] - data[:, 3] + _constant_1
    data[:, 2] = data[:, 0] - data[:, 3] + data[:, 1] + _constant_2

    return data


@pytest.fixture
def deterministic_scm_adjacency_matrix():

    adjacency_matrix = np.asarray([
        [0, 1, 1, 1],
        [0, 0, 1, 0],
        [0, 0, 0, 0],
        [0, 1, 1, 0]
    ])

    return adjacency_matrix


@pytest.fixture
def new_structural_equation():

    # X_2 = X_1 + epsilon_2, where epsilon_2 = constant_2
    def f_2(u, z):
        return u + z
    constant_2 = _constant_2
    epsilon_2 = randint(low=constant_2, high=(constant_2 + 1))
    equation_2 = StructuralEquation(2, [1], epsilon_2, f_2)

    return equation_2


@pytest.fixture
def post_intervention_deterministic_scm_data_generation_function(
        deterministic_scm, nb_samples):

    data = np.zeros(shape=(nb_samples, deterministic_scm.nb_var))
    data[:, 0] = _constant_0
    data[:, 3] = data[:, 0] + _constant_3
    data[:, 1] = data[:, 0] - data[:, 3] + _constant_1
    data[:, 2] = data[:, 1] + _constant_2

    return data


@pytest.fixture
def post_intervention_deterministic_scm_adjacency_matrix():

    adjacency_matrix = np.asarray([
        [0, 1, 0, 1],
        [0, 0, 1, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0]
    ])

    return adjacency_matrix


@pytest.fixture
def invalid_new_structural_equation():

    # X_4 = X_1 + epsilon_4, where epsilon_4 = 0
    def f_4(u, z):
        return u + z
    constant_4 = 0
    epsilon_4 = randint(low=constant_4, high=(constant_4 + 1))
    equation_4 = StructuralEquation(4, [1], epsilon_4, f_4)

    return equation_4


def test_correct_functionals(deterministic_scm, nb_samples,
                             deterministic_scm_data_generation_function):

    actual_data = deterministic_scm.generate_data(nb_samples).values

    expected_data = deterministic_scm_data_generation_function

    assert np.equal(actual_data, expected_data).all()


def test_correct_adjacency_matrix(deterministic_scm,
                                  deterministic_scm_adjacency_matrix):

    actual_adjacency_matrix = deterministic_scm.adjacency_matrix()

    assert np.equal(actual_adjacency_matrix,
                    deterministic_scm_adjacency_matrix).all()


def test_perform_intervention(
        deterministic_scm, new_structural_equation, nb_samples,
        post_intervention_deterministic_scm_data_generation_function,
        post_intervention_deterministic_scm_adjacency_matrix):

    new_scm = deterministic_scm.perform_intervention(new_structural_equation)

    # Check the data generated is what is expected
    actual_data = new_scm.generate_data(nb_samples).values
    expected_data = post_intervention_deterministic_scm_data_generation_function

    generated_data_correct = np.equal(actual_data, expected_data).all()

    # Check that the new adjacency matrix would be what is expected
    actual_adjacency_matrix = new_scm.adjacency_matrix()

    adjacency_matrix_correct = np.equal(
        actual_adjacency_matrix,
        post_intervention_deterministic_scm_adjacency_matrix).all()

    assert generated_data_correct and adjacency_matrix_correct


def test_invalid_intervention(deterministic_scm,
                              invalid_new_structural_equation):

    with pytest.raises(InvalidIntervention):

        deterministic_scm.perform_intervention(invalid_new_structural_equation)



import numpy as np

from StructuralCausalModels.graph import Graph
from StructuralCausalModels.graph import InvalidAdjacencyMatrix


class DirectedGraph(Graph):

    def __init__(self, name, adjacency_matrix):

        if not DirectedGraph.validate_directed_graph_adjacency_matrix(
                adjacency_matrix):
            msg = 'Adjacency matrix provided not valid for a directed graph.'
            raise InvalidAdjacencyMatrix(msg)

        super().__init__(name, adjacency_matrix)

    @staticmethod
    def validate_directed_graph_adjacency_matrix(matrix):
        """
        Checks that a matrix is a valid adjacency matrix for a directed graph.

        Parameters
        ----------
        matrix : array_like
            The matrix to check.

        Returns
        -------
        bool
            Whether the matrix to check is a valid adjacency matrix for a directed
            graph.
        """
        if not Graph.validate_binary_matrix(matrix):
            return False

        # Check that there are no self-loops
        no_self_loops = np.asarray([matrix[i, i] == 0 for i in
                                    range(matrix.shape[0])])
        if not np.all(no_self_loops):
            return False

        # Check that there are no undirected edges
        for i in range(matrix.shape[0]):
            for j in range(i):
                if matrix[i, j] == 1 and matrix[j, i] == 1:
                    return False

        return True



import numpy as np
import pandas as pd

from StructuralCausalModels.dag import DirectedAcyclicGraph


class CyclicityWarning(UserWarning):

    pass


class InvalidIntervention(Exception):

    pass


# TODO document
# TODO allow cycles in init yes/no ?
# TODO validation of attributes
# TODO add string representation of Structural Causal Model
class StructuralCausalModel:

    def __init__(self, name, nb_var, structural_equations):
        self.name = name
        self.nb_var = nb_var
        self.structural_equations = structural_equations

    def generate_data(self, nb_samples):

        index = range(nb_samples)
        columns = [i for i in range(self.nb_var)]
        data = pd.DataFrame(
            np.empty((nb_samples, self.nb_var,)),
            index=index,
            columns=columns
        )
        for structural_equation in self.structural_equations:
            data = structural_equation.generate_data(data)

        return data

    def adjacency_matrix(self):

        nb_nodes = len(self.structural_equations)
        adjacency_matrix = np.zeros(shape=(nb_nodes, nb_nodes))

        for structural_equation in self.structural_equations:

            parents = structural_equation.indices_rhs
            child = structural_equation.index_lhs

            for parent in parents:

                adjacency_matrix[parent, child] = 1

        adjacency_matrix = adjacency_matrix.astype(int)

        return adjacency_matrix

    def check_no_cycles(self, atol):

        adjacency_matrix = self.adjacency_matrix()

        if not DirectedAcyclicGraph.validate_dag_adjacency_matrix(
                adjacency_matrix, atol):

            msg = "The SCM defined is very likely to be cyclic. Beware !"

            raise CyclicityWarning(msg)

    def perform_intervention(self, new_structural_equation):

        target_node = new_structural_equation.index_lhs
        nodes_and_indices = dict()
        for i in range(len(self.structural_equations)):
            nodes_and_indices[self.structural_equations[i].index_lhs] = i
        existing_nodes = nodes_and_indices.keys()

        if target_node not in existing_nodes:
            # We do not allow interventions to create new (i.e. non previously
            # existing) nodes in the (DAG associated to the) SCM
            msg = "The target node does not exist in the SCM !"
            raise InvalidIntervention(msg)

        index_target_node = nodes_and_indices[target_node]
        self.structural_equations[index_target_node] = new_structural_equation

        return self



import pytest
import numpy as np

from StructuralCausalModels.graph import Graph


@pytest.mark.parametrize(
    "matrix,expected",
    [
        (np.asarray([]), False),
        (np.asarray([1, 0, 1]), False),
        (np.asarray([[1, 0, 1], [1, 0, 0]]), False),
        (np.asarray([[1, 1], [1, 1]]), True),
        (np.asarray([[0, 0, 0], [0, 0, 0], [0, 0, 0]]), True),
        (np.asarray([[1, 0, 0], [1, 2, 1], [1, 0, 1]]), False),
    ]
)
def test_validate_adjacency_matrix(matrix, expected):

    assert Graph.validate_binary_matrix(matrix) == expected


import pytest
import numpy as np

from scipy.stats import randint

from StructuralCausalModels.linear_structural_causal_model import LinearStructuralCausalModel, \
    InvalidWeightedAdjacencyMatrix, InvalidNumberOfExogenousVariables


_constant_0 = 1
_constant_1 = 0
_constant_2 = 100
_constant_3 = 0


@pytest.fixture
def nb_samples():

    return 1000


@pytest.fixture
def matrix_coefficients():
    """
    The matrix of coefficients from which to build the linear SCM (i.e. its
    weighted adjacency matrix) modulo the exogenous variables.
    """
    mat = np.asarray([
        [0, -2, 4, 2],
        [0, 0, -8, 0],
        [0, 0, 0, 0],
        [0, 0.5, 1.5, 0]
    ])

    return mat


@pytest.fixture
def causal_order():

    return [0, 3, 1, 2]


@pytest.fixture
def exogenous_variables():

    exogenous_variables = [randint(low=_constant_0, high=(_constant_0 + 1)),
                           randint(low=_constant_1, high=(_constant_1 + 1)),
                           randint(low=_constant_2, high=(_constant_2 + 1)),
                           randint(low=_constant_3, high=(_constant_3 + 1))]

    return exogenous_variables


@pytest.fixture
def linear_scm_data_generation_function(nb_samples, matrix_coefficients,
                                        exogenous_variables):
    data = np.zeros(shape=(nb_samples, matrix_coefficients.shape[0]))
    data[:, 0] = _constant_0
    data[:, 3] = matrix_coefficients[0, 3]*data[:, 0] + _constant_3
    data[:, 1] = (matrix_coefficients[0, 1]*data[:, 0] +
                  matrix_coefficients[3, 1]*data[:, 3] + _constant_1)
    data[:, 2] = (matrix_coefficients[0, 2]*data[:, 0] +
                  matrix_coefficients[3, 2]*data[:, 3] +
                  matrix_coefficients[1, 2]*data[:, 1] + _constant_2)

    return data


@pytest.fixture
def non_directed_graph_weighted_adjacency_matrix():

    mat = np.asarray([
        [0, -2, 4, 2],
        [1, 0, -8, 0],
        [0, 0, 0, 0],
        [0, 0.5, 1.5, 0]
    ])

    return mat


def test_create_from_coefficient_matrix(nb_samples, matrix_coefficients,
                                        causal_order, exogenous_variables,
                                        linear_scm_data_generation_function):

    linear_scm = LinearStructuralCausalModel.create_from_coefficient_matrix(
        name='test linear scm',
        matrix=matrix_coefficients,
        causal_order=causal_order,
        exogenous_variables=exogenous_variables)

    actual_data = linear_scm.generate_data(nb_samples).values

    expected_data = linear_scm_data_generation_function

    assert np.equal(actual_data, expected_data).all()


def test_non_directed_graph_weighted_adjacency_matrix_crashes(
        non_directed_graph_weighted_adjacency_matrix, causal_order,
        exogenous_variables):

    with pytest.raises(InvalidWeightedAdjacencyMatrix):
        LinearStructuralCausalModel.create_from_coefficient_matrix(
            name='test linear scm',
            matrix=non_directed_graph_weighted_adjacency_matrix,
            causal_order=causal_order,
            exogenous_variables=exogenous_variables)


def test_incorrect_number_of_exogenous_variables_crashes(matrix_coefficients,
                                                         causal_order,
                                                         exogenous_variables):

    with pytest.raises(InvalidNumberOfExogenousVariables):
        LinearStructuralCausalModel.create_from_coefficient_matrix(
            name='test linear scm',
            matrix=matrix_coefficients,
            causal_order=causal_order,
            exogenous_variables=exogenous_variables[:-1])



import numpy as np


class InvalidAdjacencyMatrix(Exception):
    pass


class Graph:

    def __init__(self, name, adjacency_matrix):

        if not Graph.validate_binary_matrix(adjacency_matrix):
            msg = 'Adjacency matrix provided not valid.'
            raise InvalidAdjacencyMatrix(msg)

        self.name = name
        self.adjacency_matrix = adjacency_matrix

    @staticmethod
    def validate_binary_matrix(matrix):
        """
        Checks that a matrix is an adjacency matrix (i.e. a square matrix containing
        only 0's and 1's).

        Parameters
        ----------
        matrix : array_like
            The matrix to check.

        Returns
        -------
        bool
            Whether the matrix is a valid adjacency matrix.
        """
        if len(matrix.shape) != 2:
            return False

        if matrix.shape[0] != matrix.shape[1]:
            return False

        if not np.all(np.logical_or(matrix == 1, matrix == 0)):
            return False

        return True
# def validate_binary_matrix(matrix):
#     """
#     Checks that a matrix is an adjacency matrix (i.e. a square matrix containing
#     only 0's and 1's).
#
#     Parameters
#     ----------
#     matrix : array_like
#         The matrix to check.
#
#     Returns
#     -------
#     bool
#         Whether the matrix is a valid adjacency matrix.
#     """
#     if len(matrix.shape) != 2:
#
#         return False
#
#     if matrix.shape[0] != matrix.shape[1]:
#
#         return False
#
#     if not np.all(np.logical_or(matrix == 1, matrix == 0)):
#
#         return False
#
#     return True


import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

setup(
    name='StructuralCausalModels',
    version='2.0.0',
    description='A Python package for Structural Causal Models.',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/Black-Swan-ICL/PySCMs',
    author='K. M-H',
    author_email='kmh.pro@protonmail.com',
    license='MIT',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Intended Audience :: Science/Research',
    ],
    packages=find_packages(),
    install_requires=['numpy', 'scipy', 'pandas', 'pytest']
)



import numpy as np

from StructuralCausalModels.directed_graph import DirectedGraph
from StructuralCausalModels.structural_equation import StructuralEquation
from StructuralCausalModels.structural_causal_model import StructuralCausalModel


class InvalidWeightedAdjacencyMatrix(Exception):

    pass


class InvalidNumberOfExogenousVariables(Exception):

    pass


class LinearStructuralCausalModel(StructuralCausalModel):

    def __init__(self, name, nb_var, structural_equations):
        super().__init__(name, nb_var, structural_equations)

    @staticmethod
    def create_from_coefficient_matrix(name, matrix, causal_order,
                                       exogenous_variables):

        # Check that the matrix is a valid weighted adjacency matrix for a
        # directed graph
        binarised_matrix = matrix.copy()
        binarised_matrix[binarised_matrix != 0] = 1
        if not DirectedGraph.validate_directed_graph_adjacency_matrix(
                binarised_matrix):
            msg = "The graph induced by the matrix is not a directed graph !"
            raise InvalidWeightedAdjacencyMatrix(msg)

        # Check the number of exogenous variables provided is correct
        nb_var = len(exogenous_variables)
        m = matrix.shape[0]
        if nb_var != m:
            msg = f"{nb_var} exogenous variables provided. Exactly {m} needed !"
            raise InvalidNumberOfExogenousVariables(msg)

        structural_equations = []

        for i in causal_order:

            index_lhs = i
            indices_rhs = sorted(np.where(matrix[:, i] != 0)[0].tolist())
            exogenous_variable = exogenous_variables[i]

            def generate_partial_function(index_lhs, indices_rhs):

                def func(u, *inputs):
                    if not inputs:
                        return u
                    else:
                        res = u.astype(float)
                        for j in range(len(indices_rhs)):
                            # TODO are we sure the right coeff will always go
                            #  with the right input ?
                            res += matrix[indices_rhs[j], index_lhs] * inputs[j]
                        return res

                return func

            f = generate_partial_function(index_lhs, indices_rhs)

            structural_equation = StructuralEquation(
                index_lhs=index_lhs,
                indices_rhs=indices_rhs,
                exogenous_variable=exogenous_variable,
                function=f
            )
            structural_equations.append(structural_equation)

        linear_scm = LinearStructuralCausalModel(
            name=name,
            nb_var=m,
            structural_equations=structural_equations
        )

        return linear_scm



# TODO document
# TODO add string representation of Structural Equation
class StructuralEquation:

    def __init__(self, index_lhs, indices_rhs, exogenous_variable, function):
        self.index_lhs = index_lhs
        self.indices_rhs = indices_rhs
        self.exogenous_variable = exogenous_variable
        self.function = function

    def generate_data(self, data):
        sample_size = data.shape[0]
        inputs = [data.loc[:, i].values for i in self.indices_rhs]
        data.loc[:, self.index_lhs] = self.function(
            self.exogenous_variable.rvs(size=sample_size),
            *inputs
        )

        return data
